import asyncio
from dataclasses import dataclass

from pydantic import BaseModel
from koil.decorators import koilable
from rath.errors import NotConnectedError
from rath.links.base import TerminatingLink
from typing import (
    AsyncIterator,
    Dict,
    Any,
    Iterator,
    Optional,
)
from rath.operation import GraphQLResult, opify
from contextvars import ContextVar


current_rath = ContextVar("current_rath")


@koilable(add_connectors=True)
@dataclass
class Rath:
    link: Optional[TerminatingLink] = None
    set_context: bool = True

    _connected = False

    async def aexecute(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        timeout=None,
        **kwargs,
    ) -> GraphQLResult:
        if not self._connected:
            raise NotConnectedError(
                "Rath is not connected. Please use `async with Rath(...) as rath` or use `await rath.aconnect() before`"
            )
        op = opify(query, variables, headers, operation_name, **kwargs)

        if timeout:
            return await asyncio.wait_for(self.link.aquery(op), timeout)

        return await self.link.aquery(op, **kwargs)

    def execute(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> GraphQLResult:
        if not self._connected:
            raise NotConnectedError(
                "Rath is not connected. Please use `with Rath(...) as rath` or use `rath.connect() before`"
            )
        op = opify(query, variables, headers, operation_name, **kwargs)
        return self.link.query(op, **kwargs)

    def subscribe(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> Iterator[GraphQLResult]:
        if not self._connected:
            raise NotConnectedError(
                "Rath is not connected. Please use `with Rath(...) as rath` or use `rath.connect() before`"
            )
        op = opify(query, variables, headers, operation_name, **kwargs)
        return self.link.subscribe(op, **kwargs)

    async def asubscribe(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> AsyncIterator[GraphQLResult]:
        if not self._connected:
            raise NotConnectedError(
                "Rath is not connected. Please use `async with Rath(...) as rath` or use `await rath.aconnect() before`"
            )
        op = opify(query, variables, headers, operation_name, **kwargs)
        async for res in self.link.asubscribe(op, **kwargs):
            yield res

    async def __aenter__(self):
        self.link(self)
        if self.set_context:
            current_rath.set(self)
        await self.link.__aenter__()
        self._connected = True
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.link.__aexit__(*args, **kwargs)
        self._connected = False
        self.link(None)
        if self.set_context:
            current_rath.set(None)

    def __enter__(self) -> "Rath":
        ...



from dataclasses import dataclass, field
from typing import Any, AsyncIterator, Awaitable, Callable, Coroutine
from rath.links.base import AsyncContinuationLink
from rath.operation import GraphQLResult, Operation
from rath.links.errors import AuthenticationError


async def fake_loader():
    raise Exception("No Token loader specified")


@dataclass
class AuthTokenLink(AsyncContinuationLink):

    token_loader: Callable[[], Awaitable[str]] = field(
        default_factory=lambda: fake_loader
    )
    maximum_refresh_attempts: int = 2
    load_token_on_connect: bool = True

    _token = None

    async def __aenter__(self):
        if self.load_token_on_connect:
            await self.reload_token()

    async def reload_token(self) -> None:
        self.token = await self.token_loader()
        return self.token

    async def aquery(self, operation: Operation, retry=0, **kwargs) -> GraphQLResult:
        operation.context.headers["Authorization"] = f"Bearer {self.token}"
        if not self.token:
            await self.reload_token()
        try:

            return await self.next.aquery(operation, **kwargs)
        except AuthenticationError as e:
            retry = retry + 1
            if retry > self.maximum_refresh_attempts:
                raise AuthenticationError("Maximum refresh attempts reached") from e

            self.token = await self.reload_token()
            return await self.aquery(operation, retry=retry, **kwargs)

    async def asubscribe(
        self, operation: Operation, retry=0, **kwargs
    ) -> AsyncIterator[GraphQLResult]:
        if retry >= self.maximum_refresh_attempts:
            raise AuthenticationError("Maximum refresh attempts reached")

        operation.context.headers["Authorization"] = f"Bearer {self.token}"
        try:
            async for result in self.next.asubscribe(operation, **kwargs):
                yield result

        except AuthenticationError as e:
            async for result in self.asubscribe(operation, retry=retry + 1, **kwargs):
                yield result



import asyncio
from concurrent.futures import ThreadPoolExecutor
from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from koil import unkoil, unkoil_gen, Koil
from koil.helpers import unkoil_gen


class SwitchAsyncLink(ContinuationLink):
    """Swithc Async Link

    Async switches allow to use async resolvers in an sync context.
    It uses the koil library to facilitate the switch.

    Koil can use unstateful links (no "async with" logic), by simply calling
    asyncio.run(). Or spin up a thread with the async logic for stateful links
    (with "async with" logic). It also takes care of stoppin the thread when
    the link is disconnected.

    This allows us to use the same logic for both sync and async links, where
    a connection is required.

    Check the koil documentation to learn more about how koil does it in detail.

    Args:
        ContinuationLink (_type_): _description_
    """

    def __init__(self, **koilparams):
        super().__init__()

    async def aquery(self, operation: Operation, **kwargs) -> GraphQLResult:
        return await self.next.aquery(operation)

    async def asubscribe(self, operation: Operation, **kwargs) -> GraphQLResult:
        async for result in self.next.asubscribe(operation):
            yield result

    def query(self, operation: Operation, **kwargs) -> GraphQLResult:
        return unkoil(self.next.aquery, operation, **kwargs)

    def subscribe(self, operation: Operation, **kwargs) -> GraphQLResult:
        return unkoil_gen(self.next.asubscribe, operation, **kwargs)



from typing import AsyncIterator, Iterator
from rath.links.errors import LinkNotConnectedError
from rath.operation import GraphQLResult, Operation
from koil import unkoil, unkoil_gen


class Link:
    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, *args, **kwargs) -> None:
        pass

    async def aquery(self, operation: Operation, **kwargs) -> GraphQLResult:
        raise NotImplementedError(
            f"Please overwrite the aquery method in {self.__class__.__name__}"
        )

    async def asubscribe(
        self, operation: Operation, **kwargs
    ) -> AsyncIterator[GraphQLResult]:
        raise NotImplementedError(
            f"Please overwrite the asubscribe method in {self.__class__.__name__}"
        )

    def query(self, operation: Operation, **kwargs) -> GraphQLResult:
        raise NotImplementedError(
            f"Please overwrite the query method in {self.__class__.__name__}"
        )

    def subscribe(self, operation: Operation, **kwargs) -> Iterator[GraphQLResult]:
        raise NotImplementedError(
            f"Please overwrite the subscribe method in {self.__class__.__name__}"
        )


class TerminatingLink(Link):
    def __call__(self, rath):
        self.rath = rath
        return self


class AsyncTerminatingLink(TerminatingLink):
    def aquery(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError("Your Async Transport needs to overwrite this method")

    def asubscribe(self, operation: Operation) -> Iterator[GraphQLResult]:
        raise NotImplementedError("Your Async Transport needs to overwrite this method")

    def query(self, operation: Operation, **kwargs) -> GraphQLResult:
        return unkoil(self.aquery, operation, **kwargs)

    def subscribe(self, operation: Operation, **kwargs) -> GraphQLResult:
        return unkoil_gen(self.asubscribe, operation, **kwargs)


class SyncTerminatingLink(TerminatingLink):
    async def aquery(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Sync Terminating link does not support asynchronous queries. Please compose together with a context switch link"
        )

    async def asubscribe(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        raise NotImplementedError(
            "Sync Terminating link does not support asynchronous queries. Please compose together with a context switch link"
        )


class ContinuationLink(Link):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._next = None

    def __call__(self, rath, next: Link):
        self.rath = rath
        self._next = next
        return self

    @property
    def next(self):
        if self._next is None:
            raise LinkNotConnectedError(
                "Next link is not set. This means we were never connected. Please use rath in an async context"
            )
        return self._next

    async def aquery(self, operation: Operation, **kwargs) -> GraphQLResult:
        return await self.next.aquery(operation, **kwargs)

    async def asubscribe(self, operation: Operation, **kwargs) -> GraphQLResult:
        async for x in self.next.asubscribe(operation, **kwargs):
            yield x

    def query(self, operation: Operation, **kwargs) -> GraphQLResult:
        return self.next.query(operation, **kwargs)

    def subscribe(self, operation: Operation, **kwargs) -> GraphQLResult:
        return self.next.subscribe(operation, **kwargs)


class AsyncContinuationLink(ContinuationLink):
    def query(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Async Terminating link does not support syncrhonous queries. Please compose together with a context switch link"
        )

    def subscribe(self, operation: Operation) -> Iterator[GraphQLResult]:
        raise NotImplementedError(
            "Async Terminating link does not support syncrhonous queries. Please compose together with a context switch link"
        )


class SyncContinuationLink(ContinuationLink):
    async def aquery(self, operation: Operation) -> GraphQLResult:
        raise NotImplementedError(
            "Sync Terminating link does not support asynchronous queries. Please compose together with a context switch link"
        )

    async def asubscribe(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        raise NotImplementedError(
            "Sync Terminating link does not support asynchronous queries. Please compose together with a context switch link"
        )



from dataclasses import dataclass
from typing import Callable, List
from rath.operation import Operation
from rath.links.base import TerminatingLink


@dataclass
class SplitLink(TerminatingLink):
    left: TerminatingLink
    right: TerminatingLink
    split: Callable[[Operation], bool]

    def __call__(self, rath):
        self.left(rath)
        self.right(rath)

    async def aquery(self, operation: Operation, **kwargs) -> Operation:
        future = (
            self.left.aquery(operation, **kwargs)
            if self.split(operation)
            else self.right.aquery(operation, **kwargs)
        )
        return await future

    async def asubscribe(self, operation: Operation, **kwargs) -> Operation:
        iterator = (
            self.left.asubscribe(operation, **kwargs)
            if self.split(operation)
            else self.right.asubscribe(operation, **kwargs)
        )

        async for res in iterator:
            yield res

    def query(self, operation: Operation, **kwargs) -> Operation:
        future = (
            self.left.query(operation, **kwargs)
            if self.split(operation)
            else self.right.query(operation, **kwargs)
        )

        return future

    def subscribe(self, operation: Operation, **kwargs) -> Operation:
        iterator = (
            self.left.subscribe(operation, **kwargs)
            if self.split(operation)
            else self.right.subscribe(operation, **kwargs)
        )

        return iterator

    async def __aenter__(self):
        await self.left.__aenter__()
        await self.right.__aenter__()

    async def __aexit__(self, *args, **kwargs):
        await self.left.__aexit__(*args, **kwargs)
        await self.right.__aexit__(*args, **kwargs)


def split(
    left: TerminatingLink, right: TerminatingLink, split: Callable[[Operation], bool]
):
    """
    Splits a Link into two paths. Acording to a predicate function. If predicate returns
    true, the operation is sent to the left path, otherwise to the right path.
    """
    return SplitLink(left, right, split)
